@@ -1,7 +1,9 @@
 """Main RequestSession module."""
+
 import re
 import time
 from collections import namedtuple
+from copy import deepcopy
 from typing import List  # pylint: disable=unused-import
 from typing import Any, Callable, Dict, Optional, Tuple, Union
 
@@ -40,6 +42,9 @@ class RequestSession(object):
         server error. Defaults to 0.
     :param bool verbose_logging: (optional) If true, add request's parameters to event
         being logged. Defaults to ``False``.
+    :param tuple request_param_logging_blacklist: (optional)
+        Request params keys that won't be logged with verbose_logging.
+        e.g. auth or header keys might contain secrets. Defaults to ``("auth", "headers")``.
     :param str request_category: (optional) Name of the event. ``request_category`` has
         to passed to the object or as an argument when calling some HTTP method.
     :param bool raise_for_status: (optional) Raise an exception in case of an error.
@@ -49,6 +54,8 @@ class RequestSession(object):
         ``requests.Session`` to be used to make the HTTP requests.
     :param Ddtrace ddtrace: (optional) DataDog function to be used to trace, track and
         send metrics for individual HTTP requests.
+    :param bool datadog_errors: (optional) If true, errors will be sent to DataDog.
+        Defaults to ``False``.
     :param str ddtrace_service_name: (optional) Service name passed to DataDog DdTrace. Set to "booking_api_requests"
         by default.
     :param Statsd statsd: (optional) Datadog module to log metrics.
@@ -72,12 +79,14 @@ class RequestSession(object):
         verify=True,  # type: Union[bool, str]
         max_retries=0,  # type: int
         verbose_logging=False,  # type: bool
+        request_param_logging_blacklist=None,  # type: Optional[Tuple[str]]
         headers=None,  # type: Optional[Dict]
         request_category=None,  # type: Optional[str]
         raise_for_status=True,  # type: bool
         user_agent=None,  # type: Optional[str]
         user_agent_components=None,  # type: Optional[UserAgentComponents]
         ddtrace=None,  # type: Optional[Ddtrace]
+        datadog_errors=False,  # type: bool
         statsd=None,  # type: Optional[Statsd]
         sentry_client=None,  # type: Optional[SentryClient]
         logger=None,  # type: Optional[Callable]
@@ -92,6 +101,7 @@ class RequestSession(object):
             "log",
         ),  # type: Tuple
         ddtrace_service_name="booking_api_requests",  # type: str
+        retriable_client_errors=None,  # type: Optional[List[int]]
     ):
         # type: (...) -> None
         self.host = host
@@ -100,6 +110,10 @@ class RequestSession(object):
         self.verify = verify
         self.max_retries = max_retries
         self.verbose_logging = verbose_logging
+        self.request_param_logging_blacklist = request_param_logging_blacklist or (
+            "auth",
+            "headers",
+        )
         self.headers = headers
         self.request_category = request_category
         self.raise_for_status = raise_for_status
@@ -107,11 +121,15 @@ class RequestSession(object):
         self.user_agent_components = user_agent_components
         self.ddtrace = ddtrace
         self.ddtrace_service_name = ddtrace_service_name
+        self.datadog_errors = datadog_errors
         self.statsd = statsd
         self.sentry_client = sentry_client
         self.logger = logger
         self.log_prefix = log_prefix
         self.allowed_log_levels = allowed_log_levels
+        self.retriable_client_errors = (
+            retriable_client_errors if retriable_client_errors else [408]
+        )
 
         self.prepare_new_session()
 
@@ -147,17 +165,7 @@ class RequestSession(object):
         # type: () -> None
         """Set proper user-agent string to header according to RFC22."""
         pattern = r"^(?P<service_name>\S.+?)\/(?P<version>\S.+?) \((?P<organization>\S.+?) (?P<environment>\S.+?)\)(?: ?(?P<sys_info>.*))$"
-        string = (
-            "{service_name}/{version} ({organization} {environment}) {sys_info}".format(
-                service_name=self.user_agent_components.service_name,  # type: ignore
-                version=self.user_agent_components.version,  # type: ignore
-                organization=self.user_agent_components.organization,  # type: ignore
-                environment=self.user_agent_components.environment,  # type: ignore
-                sys_info=self.user_agent_components.sys_info  # type: ignore
-                if self.user_agent_components.sys_info  # type: ignore
-                else "",
-            ).strip()
-        )
+        string = f"{self.user_agent_components.service_name}/{self.user_agent_components.version} ({self.user_agent_components.organization} {self.user_agent_components.environment}) {self.user_agent_components.sys_info if self.user_agent_components.sys_info else ''}".strip()
         if not re.match(pattern, string):
             raise InvalidUserAgentString("Provided User-Agent string is not valid.")
         self.user_agent = string
@@ -271,7 +279,7 @@ class RequestSession(object):
         sleep_before_repeat=None,  # type: Optional[float]
         tags=None,  # type: Optional[list]
         raise_for_status=None,  # type: Optional[bool]
-        **request_kwargs  # type: Any
+        **request_kwargs,  # type: Any
     ):  # pylint: disable=too-many-statements
         # type: (...) -> Optional[requests.Response]
         r"""Run a request against a service depending on a request type.
@@ -313,7 +321,7 @@ class RequestSession(object):
         max_runs = 1 + (self.max_retries if max_retries is None else max_retries)
         run, retries_on_econnreset = 0, 0
         # this will set maximum number of retries to max_runs where econnreset retries
-        # are not counting and maximum number of of retries on econnreset is also
+        # are not counting and maximum number of retries on econnreset is also
         # set to max_runs
         while run < max_runs + retries_on_econnreset:
             run += 1
@@ -364,9 +372,11 @@ class RequestSession(object):
                         attempt=run,
                     )
 
-                if self.is_server_error(error, status_code):
+                if self.is_server_error(
+                    error, status_code
+                ) or self.retry_on_client_errors(status_code):
                     if is_econnreset_error:
-                        self.log("info", "{}.session_replace".format(request_category))
+                        self.log("info", f"{request_category}.session_replace")
                         self.remove_session()
                         self.prepare_new_session()
 
@@ -398,21 +408,28 @@ class RequestSession(object):
                 else:
                     # Client error, request is not valid and server rejected it with
                     # http 4xx, but not timeout, there is no point in retrying.
-                    if report and self.sentry_client is not None:
+                    if report:
                         response_text = self.get_response_text(response)
                         extra_data = (
                             {"response_text": response_text}
                             if response_text != ""
                             else None
                         )
-                        if callable(
-                            getattr(self.sentry_client, "capture_exception", None)
-                        ):
-                            self.sentry_client.capture_exception(extras=extra_data)
-                        elif callable(
-                            getattr(self.sentry_client, "captureException", None)
-                        ):
-                            self.sentry_client.captureException(extra=extra_data)
+
+                        if self.sentry_client is not None:
+                            if callable(
+                                getattr(self.sentry_client, "capture_exception", None)
+                            ):
+                                self.sentry_client.capture_exception(extras=extra_data)
+                            elif callable(
+                                getattr(self.sentry_client, "captureException", None)
+                            ):
+                                self.sentry_client.captureException(extra=extra_data)
+
+                        if self.datadog_errors and self.ddtrace is not None:
+                            span = self.ddtrace.tracer.current_span()
+                            if span:
+                                span.record_exception(error, extra_data)
 
                     if raise_for_status:
                         raise RequestSessionException(  # pylint: disable=raise-missing-from
@@ -442,25 +459,13 @@ class RequestSession(object):
         :param str request_category: Category for log and metric reporting.
         :return requests.Response: HTTP Response Object.
         """
-        metric_name = "{request_category}.response_time".format(
-            request_category=request_category
-        )
+        metric_name = f"{request_category}.response_time"
 
         if not self.statsd:
             return self.session.request(method=request_type, **request_params)
 
-        with self.statsd.timed(metric_name, use_ms=True, tags=tags):
-            if self.ddtrace:
-                with self.ddtrace.tracer.trace(
-                    "api.request",
-                    service=self.ddtrace_service_name,
-                    resource=request_category,
-                ):
-                    response = self.session.request(
-                        method=request_type, **request_params
-                    )
-            else:
-                response = self.session.request(method=request_type, **request_params)
+        with self.statsd.distributed(metric_name, use_ms=True, tags=tags):
+            response = self.session.request(method=request_type, **request_params)
         return response
 
     def _log_with_params(
@@ -476,7 +481,13 @@ class RequestSession(object):
         """
         extra_params = (
             {
-                "request_params": json.dumps(request_params),
+                "request_params": json.dumps(
+                    {
+                        k: v
+                        for k, v in deepcopy(request_params).items()
+                        if k not in self.request_param_logging_blacklist
+                    }
+                ),
                 "response_text": self.get_response_text(response),
             }
             if self.verbose_logging
@@ -489,7 +500,7 @@ class RequestSession(object):
             status_code=response.status_code,
             attempt=attempt,
             url=url,
-            **extra_params
+            **extra_params,
         )
 
     def sleep(self, seconds, request_category, tags):
@@ -501,9 +512,9 @@ class RequestSession(object):
         :param List[str] tags: tags for datadog
         """
         trace_name = request_category.replace(".", "_") + "_retry"
-        meta = {"request_category": request_category}
-        split_tags_and_update(meta, tags)
-        traced_sleep(trace_name, seconds, self.ddtrace, meta)  # type: ignore
+        default_tags = {"request_category": request_category}
+        split_tags_and_update(default_tags, tags)
+        traced_sleep(trace_name, seconds, self.ddtrace, default_tags)  # type: ignore
 
     def metric_increment(self, metric, request_category, tags, attempt=None):
         # type: (str, str, list, Optional[int]) -> None
@@ -516,12 +527,10 @@ class RequestSession(object):
         """
         new_tags = list(tags) if tags else []
         if attempt:
-            new_tags.append("attempt:{attempt}".format(attempt=attempt))
+            new_tags.append(f"attempt:{attempt}")
 
         if self.statsd is not None:
-            metric_name = "{metric_base}.{metric_type}".format(
-                metric_base=request_category, metric_type=metric
-            )
+            metric_name = f"{request_category}.{metric}"
             self.statsd.increment(metric_name, tags=new_tags)
 
     def log(self, level, event, **kwargs):
@@ -535,7 +544,7 @@ class RequestSession(object):
         """
         if not level in self.allowed_log_levels:
             raise AttributeError("Provided log level is not allowed.")
-        event_name = "{prefix}.{event}".format(prefix=self.log_prefix, event=event)
+        event_name = f"{self.log_prefix}.{event}"
         if self.logger is not None:
             getattr(self.logger, level)(event_name, **kwargs)
 
@@ -551,11 +560,11 @@ class RequestSession(object):
         # type: (...) -> None
         """Assign appropriate metric and log for exception.
 
-        :param requests.RequestException error: exception that occured
+        :param requests.RequestException error: exception that occurred
         :param str request_category: String describing request category.
         :param Dict request_params: Parameters used to make the HTTP call.
         :param List[str] dd_tags: Tags to increment metric with.
-        :param Union[int, None] Status_code: HTTP status code of the response.
+        :param Union[int, None] status_code: HTTP status code of the response.
         """
         tags = ["status:error"]
 
@@ -580,21 +589,27 @@ class RequestSession(object):
             tags.append("error:request_exception")
 
         if self.verbose_logging is True:
-            extra_params["request_params"] = json.dumps(request_params)
+            extra_params["request_params"] = json.dumps(
+                {
+                    k: v
+                    for k, v in deepcopy(request_params).items()
+                    if k not in self.request_param_logging_blacklist
+                }
+            )
 
         self.log(
             "exception",
-            "{}.failed".format(request_category),
+            f"{request_category}.failed",
             error_type=error_type,
             status_code=status_code,
             attempt=attempt,
-            **extra_params
+            **extra_params,
         )
 
         self.metric_increment(
             metric="request",
             request_category=request_category,
-            tags=tags + ["attempt:{}".format(attempt)],
+            tags=tags + [f"attempt:{attempt}"],
         )
 
     @staticmethod
@@ -609,11 +624,23 @@ class RequestSession(object):
         if not isinstance(error, requests.exceptions.HTTPError):
             return True
 
-        if http_code is not None and (400 <= http_code < 500 and not http_code == 408):
+        if http_code is not None and (400 <= http_code < 500):
             return False
 
         return True
 
+    def retry_on_client_errors(self, http_code):
+        # type: (Optional[int]) -> bool
+        """Decide if retry should be done even on client http error.
+
+        :param int http_code: (optional) response HTTP status code
+        :return bool: whether retry should occur
+        """
+        if http_code is not None and http_code in self.retriable_client_errors:
+            return True
+
+        return False
+
     @staticmethod
     def get_response_text(response):
         # type: (Union[requests.Response, Any]) -> str
