@@ -1,9 +1,9 @@
 """Test the main module."""
+
 import itertools
 import sys
 from typing import Any, Callable, Dict, Iterator, List, Union
 
-import httpbin as Httpbin
 import pytest
 import requests
 
@@ -19,21 +19,20 @@ from request_session.protocols import Ddtrace, SentryClient, Statsd
 from ._compat import Mock
 
 REQUEST_CATEGORY = "test"  # this request category must match the one in conftest
-INTERNAL_ERROR_MSG = "500 Server Error: INTERNAL SERVER ERROR for url:"
-TIMEOUT_ERROR_MSG = "408 Client Error: REQUEST TIMEOUT for url:"
+INTERNAL_ERROR_MSG = "500 Server Error:"
+TIMEOUT_ERROR_MSG = "408 Client Error:"
 DDTRACE_ERROR_MSG = "Ddtrace must be provided in order to report to datadog service."
 USER_AGENT_ERROR_MSG = "Provided User-Agent string is not valid."
 
 
-def test_init(mocker, httpbin):
-    # type: (Mock, Httpbin) -> None
+def test_init(mocker, test_server):
     """Test initialization of RequestSession."""
     mock_ddtrace = mocker.Mock(spec_set=Ddtrace)
-    mock_tracing_config = dict()  # type: Dict[Any, Any]
+    mock_tracing_config: dict = {}
     mock_ddtrace.config.get_from.return_value = mock_tracing_config
 
     session = RequestSession(
-        host=httpbin.url,
+        host=test_server,
         request_category=REQUEST_CATEGORY,
         max_retries=3,
         user_agent="UserAgent",
@@ -41,7 +40,7 @@ def test_init(mocker, httpbin):
         headers={},
         auth=("user", "passwd"),
     )
-    assert session.host == httpbin.url
+    assert session.host == test_server
     assert session.request_category == REQUEST_CATEGORY
     assert session.max_retries == 3
     assert session.user_agent == "UserAgent"
@@ -171,21 +170,20 @@ def test_method(request_session, method, path, expected_status):
         (505, HTTPError),
     ],
 )
-def test_raise_for_status(mocker, httpbin, status_code, raises):
-    # type: (Mock, Httpbin, int, Exception) -> None
+def test_raise_for_status(mocker, test_server, status_code, raises):
     """Test raising of an exception when rejected with 4xx."""
-    session = RequestSession(host=httpbin.url, request_category=REQUEST_CATEGORY)
+    session = RequestSession(host=test_server, request_category=REQUEST_CATEGORY)
     mock_sys = mocker.patch("request_session.utils.sys", spec_set=sys)
     mock_sys.exc_info.return_value = (HTTPError, HTTPError(), "fake_traceback")
     if raises:
         with pytest.raises(raises):
-            session.get(path="/status/{status_code}".format(status_code=status_code))
+            session.get(path=f"/status/{status_code}")
 
         if isinstance(raises, HTTPError):
             assert mock_sys.exc_info()[1].__sentry_source == "third_party"
             assert mock_sys.exc_info()[1].__sentry_pd_alert == "disabled"
     else:
-        session.get(path="/status/{status_code}".format(status_code=status_code))
+        session.get(path=f"/status/{status_code}")
 
 
 @pytest.mark.parametrize(
@@ -207,9 +205,9 @@ def test_raise_for_status(mocker, httpbin, status_code, raises):
     ],
 )
 def test_econnreset_error(
-    httpbin, mocker, exceptions, max_retries, expected_call_count
+    test_server, mocker, exceptions, max_retries, expected_call_count
 ):
-    # type: (Httpbin, Mock, Iterator[Exception], int, int) -> None
+    # type: (str, Mock, Iterator[Exception], int, int) -> None
     used_sessions = []
 
     def _prepare_new_session(self):  # type: ignore
@@ -220,7 +218,7 @@ def test_econnreset_error(
         used_sessions.append(self.session)
 
     client = RequestSession(
-        host=httpbin.url, max_retries=max_retries, request_category=REQUEST_CATEGORY
+        host=test_server, max_retries=max_retries, request_category=REQUEST_CATEGORY
     )
     mock_log = mocker.Mock(autospec=True)
     mock_exception_log_and_metrics = mocker.Mock(
@@ -239,9 +237,7 @@ def test_econnreset_error(
         client.get("/status/500")
     actual_call_count = sum(session.request.call_count for session in used_sessions)
 
-    mock_log.assert_called_with(
-        "info", "{category}.session_replace".format(category=client.request_category)
-    )
+    mock_log.assert_called_with("info", f"{client.request_category}.session_replace")
     assert mock_exception_log_and_metrics.call_count == 1
     assert actual_call_count == expected_call_count
 
@@ -305,7 +301,7 @@ def test_logging(mocker, request_session, inputs, expected):
     )
     client = request_session(max_retries=inputs["max_retries"])
     client._exception_log_and_metrics = mock_exception_log_and_metrics
-    expected["request_params"]["url"] = "{}{}".format(client.host, inputs["path"])
+    expected["request_params"]["url"] = f"{client.host}{inputs['path']}"
 
     calls = []
     for attempt in range(1, expected["call_count"] + 1):
@@ -360,19 +356,19 @@ def test_metric_increment(
 
     calls = []
     for attempt in range(1, call_count + 1):
-        metric = "{}.{}".format(client._get_request_category(), "request")
-        tags = ["status:{}".format(status)]
+        metric = f"{client._get_request_category()}.request"
+        tags = [f"status:{status}"]
         if error:
-            tags.append("error:{}".format(error))
-        calls.append(mocker.call(metric, tags=tags + ["attempt:{}".format(attempt)]))
+            tags.append(f"error:{error}")
+        calls.append(mocker.call(metric, tags=tags + [f"attempt:{attempt}"]))
 
     assert mock_statsd.increment.call_count == call_count
     mock_statsd.increment.assert_has_calls(calls)
 
 
-def test_get_request_category(httpbin):
-    # type: (Httpbin) -> None
-    client = RequestSession(host=httpbin.url)
+def test_get_request_category(test_server):
+    """Test getting request category."""
+    client = RequestSession(host=test_server)
 
     with pytest.raises(
         AttributeError, match="`request_category` is required parameter."
@@ -457,15 +453,15 @@ def test_send_request(request_session, mocker, inputs, expected):
     )
 
     assert isinstance(response, requests.Response)
-    mock_statsd.timed.assert_called_once_with(
-        "{}.response_time".format(client.request_category),
+    mock_statsd.distributed.assert_called_once_with(
+        f"{client.request_category}.response_time",
         use_ms=True,
         tags=inputs["tags"],
     )
 
 
-def test_sleep(httpbin, mocker):
-    # type: (Httpbin, Mock) -> None
+def test_sleep(test_server, mocker):
+    """Test sleep functionality."""
     seconds = 1
     tags = ["testing:sleep"]
     meta = {"request_category": REQUEST_CATEGORY, "testing": "sleep"}
@@ -473,7 +469,7 @@ def test_sleep(httpbin, mocker):
     mock_traced_sleep = mocker.patch(
         "request_session.request_session.traced_sleep", autospec=True
     )
-    client = RequestSession(host=httpbin.url, ddtrace=mock_ddtrace)
+    client = RequestSession(host=test_server, ddtrace=mock_ddtrace)
     client.sleep(seconds, REQUEST_CATEGORY, tags)
 
     mock_traced_sleep.assert_called_once_with(
@@ -608,7 +604,7 @@ def test_exception_and_log_metrics(request_session, mocker, inputs, expected):
 
     mock_log.assert_called_once_with(
         "exception",
-        "{}.failed".format(client.request_category),
+        f"{client.request_category}.failed",
         error_type=expected["error_type"],
         status_code=inputs["status_code"],
         **expected["extra_params"],
@@ -646,7 +642,7 @@ def test_reporting(request_session, mocker):
         (requests.exceptions.Timeout(), None, True),
         (requests.exceptions.HTTPError(), 400, False),
         (requests.exceptions.HTTPError(), 399, True),
-        (requests.exceptions.HTTPError(), 408, True),  # Timeout is server error
+        (requests.exceptions.HTTPError(), 408, False),
         (requests.exceptions.HTTPError(), 499, False),
         (requests.exceptions.HTTPError(), 500, True),
         (requests.exceptions.HTTPError(), None, True),
@@ -655,3 +651,22 @@ def test_reporting(request_session, mocker):
 def test_is_server_error(exception, status_code, expected):
     # type: (RequestException, Union[int, None], bool) -> None
     assert RequestSession.is_server_error(exception, status_code) == expected
+
+
+@pytest.mark.parametrize(
+    ("status_code, extended_retry_errors, expected"),
+    [
+        (None, [], False),
+        (408, [], True),  # Timeout is retried by default
+        (408, [429], False),
+        (408, [408, 429], True),
+        (200, [], False),
+        (429, [], False),
+        (429, [429], True),
+        (500, [], False),
+    ],
+)
+def test_retry_on_client_errors(status_code, extended_retry_errors, expected):
+    # type: (RequestException, Union[int, None], bool) -> None
+    session = RequestSession(retriable_client_errors=extended_retry_errors)
+    assert session.retry_on_client_errors(status_code) == expected
